@@ -1,8 +1,7 @@
-#!/usr/bin/perl -w
-
 package HTTPTest;
 
 use strict;
+use warnings;
 
 use HTTPServer;
 use WgetTest;
@@ -14,7 +13,7 @@ my $VERSION = 0.01;
 {
     my %_attr_data = ( # DEFAULT
     );
-    
+
     sub _default_for
     {
         my ($self, $attr) = @_;
@@ -22,25 +21,34 @@ my $VERSION = 0.01;
         return $self->SUPER::_default_for($attr);
     }
 
-    sub _standard_keys 
+    sub _standard_keys
     {
         my ($self) = @_;
         ($self->SUPER::_standard_keys(), keys %_attr_data);
     }
 }
-    
 
-sub _setup_server {}
+
+sub _setup_server {
+    my $self = shift;
+    $self->{_server} = HTTPServer->new (LocalAddr => 'localhost',
+                                        ReuseAddr => 1)
+                                    or die "Cannot create server!!!";
+}
 
 
 sub _launch_server {
     my $self = shift;
     my $synch_func = shift;
 
-    my $server = HTTPServer->new (LocalAddr => 'localhost',
-                                  LocalPort => '8080',
-                                  ReuseAddr => 1) or die "Cannot create server!!!";
-    $server->run ($self->{_input}, $synch_func);
+    $self->{_server}->run ($self->{_input}, $synch_func);
+}
+
+sub _substitute_port {
+    my $self = shift;
+    my $ret = shift;
+    $ret =~ s/{{port}}/$self->{_server}->sockport/eg;
+    return $ret;
 }
 
 1;
